import numpy as np
import os
from scipy.interpolate import griddata
from scipy.sparse import coo_matrix
import EMpy_gpu
import collections
from emepy.materials import *
from typing import Callable
from shapely.geometry import Polygon, Point


def polygon_to_n_2D(
    polygon: "Polygon", x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4
) -> "np.ndarray":
    # Create grid
    xx, yy = np.meshgrid(x, y)
    n = np.zeros(xx.shape)[:-1, :-1].T

    # Apply subpixel
    xlower, xupper = (x[:-1], x[1:])
    zlower, zupper = (y[:-1], y[1:])
    for i, xp in enumerate(zip(xlower, xupper)):
        for j, zp in enumerate(zip(zlower, zupper)):

            # Upper and lower points
            xl, xu = xp
            zl, zu = zp
            total_area = (xu - xl) * (zu - zl)

            # Create polygon of the pixel
            pixel_poly = Polygon([(xl, zl), (xl, zu), (xu, zu), (xu, zl)])

            # Get overlapping area
            overlapping_area = 0
            if pixel_poly.intersects(polygon):
                overlapping_area = pixel_poly.intersection(polygon).area

            # Calculate effective index
            if subpixel:
                n[i, j] = (
                    overlapping_area / total_area * core_index + (1 - overlapping_area / total_area) * cladding_index
                )
            elif overlapping_area:
                n[i, j] = core_index
            else:
                n[i, j] = cladding_index

    return n


def vertices_to_n(vertices: list, x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4) -> "np.ndarray":
    """
    Takes vertices of a polygon and maps it to a grid using or not using subpixel smoothing
    """

    polygon = Polygon(vertices)
    return polygon_to_n_2D(polygon, x, y, subpixel, core_index, cladding_index)


def rectangle_to_n(
    center: tuple, width: float, thickness: float, x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4
) -> "np.ndarray":

    xc, yc = center
    vertices = [
        (xc - width / 2, yc - thickness / 2),
        (xc - width / 2, yc + thickness / 2),
        (xc + width / 2, yc + thickness / 2),
        (xc + width / 2, yc - thickness / 2),
    ]
    return vertices_to_n(vertices, x, y, subpixel, core_index, cladding_index)


def circle_to_n(
    center: tuple, radius: float, x: list, y: list, subpixel=True, core_index=3.4, cladding_index=1.4
) -> "np.ndarray":

    polygon = Point(*center).buffer(radius)
    return polygon_to_n_2D(polygon, x, y, subpixel, core_index, cladding_index)


class get_epsfunc(object):
    """Callable class for getting epsilon on a grid"""

    def __init__(
        self,
        width: float,
        thickness: float,
        cladding_width: float,
        cladding_thickness: float,
        core_index: float,
        cladding_index: float,
        compute: bool = False,
        profile: "np.ndarray" = None,
        nx: int = None,
        ny: int = None,
    ):
        """Returns the epsfunc for given parameters for a rectangular waveguide

        Parameters
        ----------
        width: float
            the width of the geometry
        thickness: float
            the thickess of the geometry
        cladding_width: float
            the width of the surrounding cladding (note this is total width and should be > core width)
        cladding_thickness: float
            the thickness of the surrounding cladding (note this is total thickness and should be > core thickness)
        core_index: float
            refractive index of the core
        cladding_index: float
            refractive index of the cladding
        compute: bool = False
            if true, will not place rectangle at center. This should only be necessary for compute_other_fields
        profile: "np.ndarray" = None
            the refractive index profile (note if providing a width, this should be left None)
        nx: int = None
            number of points in the x direction
        ny: int = None
            number of points in the y direction
        subpixel: bool = True
            if true, will use subpixel smoothing, assuming asking for a waveguide cross section and not providing an index map (recommended)

        Returns
        -------
        function
            an epsfunc function that takes an x,y grid arrays and returns the refractive index profile
        """
        self.width = width
        self.thickness = thickness
        self.cladding = cladding_width
        self.cladding = cladding_thickness
        self.core_index = core_index
        self.cladding_index = cladding_index
        self.compute = compute
        self.profile = profile
        self.nx = nx
        self.ny = ny

        if (width is not None) and (thickness is not None):
            self.epsfunction = self.epsfunc_2D_1

        elif (width is None) and (thickness is not None) and (profile is not None):
            self.epsfunction = self.epsfunc_2D_2

        elif (width is None) and (thickness is None) and (profile is not None):
            self.epsfunction = self.epsfunc_2D_3

        elif (thickness is None) and (width is not None):
            self.epsfunction = self.epsfunc_1D_1

        elif (thickness is None) and (width is None) and (profile is not None):
            self.epsfunction = self.epsfunc_1D_2

        else:
            raise Exception(
                "Need to provide width & thickness, or 1D profile and thickness, or 2D profile, or width for 1D, or 1D profile for 1D"
            )

    def __call__(self, x_, y_):
        return self.epsfunction(x_, y_)

    # Case 1 : width and thickness are defined
    def epsfunc_2D_1(self, x_, y_):
        """Return a matrix describing a 2d material.

        Parameters
        ----------
        x_: numpy array
            x values
        y_: numpy array
            y values

        Returns
        -------
        numpy array
            2d-matrix
        """
        xx, yy = np.meshgrid(x_, y_)
        if self.compute:
            n = np.where(
                (np.abs(np.real(xx.T) - self.cladding_width * 0.5) <= self.width * 0.5)
                * (np.abs(np.real(yy.T) - self.cladding_thickness * 0.5) <= self.thickness * 0.5),
                self.core_index ** 2 + 0j,
                self.cladding_index ** 2 + 0j,
            )
        else:
            n = np.where(
                (np.abs(np.real(xx.T)) <= self.width * 0.5) * (np.abs(np.real(yy.T)) <= self.thickness * 0.5),
                self.core_index ** 2 + 0j,
                self.cladding_index ** 2 + 0j,
            )
        return n

    # Case 2 : thickness and 1D n is defined
    def epsfunc_2D_2(self, x_, y_):

        n = self.profile
        xx, yy = np.meshgrid(np.real(x_), np.real(y_))
        n = np.interp(np.real(x_), np.real(self.nx), n).astype(complex)
        n = np.repeat(n, len(y_)).reshape((len(n), len(y_)))
        n = np.where((np.abs(np.real(yy.T)) <= self.thickness * 0.5), n, self.cladding_index + 0j) ** 2
        return n

    # Case 3 : 2D n is defined
    def epsfunc_2D_3(self, x_, y_):

        xxn, yyn = np.meshgrid(np.real(self.nx), np.real(self.ny))
        points = np.array((xxn.flatten(), yyn.flatten())).T
        n = self.profile.flatten()
        xx, yy = np.meshgrid(np.real(x_), np.real(y_))
        n_real = griddata(points, np.real(n), (xx, yy))
        n_imag = griddata(points, np.imag(n), (xx, yy))
        n = (n_real + 1j * n_imag) ** 2
        return n

    # Case 4: width only
    def epsfunc_1D_1(self, x_, y_):

        n = np.where(
            (np.abs(np.real(x_)) <= self.width * 0.5), self.core_index ** 2 + 0j, self.cladding_index ** 2 + 0j
        ).reshape(len(x_), 1)

        return n

    # Case 5: 1D n only
    def epsfunc_1D_2(self, x_, y_):

        n = np.interp(np.real(x_), np.real(self.nx), self.profile).astype(complex).reshape(len(x_), 1)
        return n


def get_isotropic_epsfunc(
    epsfunc_xx, epsfunc_yy, epsfunc_zz, epsfunc_xy=None, epsfunc_yx=None
) -> Callable[["np.ndarray", "np.ndarray"], "np.ndarray"]:
    """Returns an epsfunction for an isotropic medium"""

    def epsfunc_iso(x_, y_):
        xx = epsfunc_xx(x_, y_)
        yy = epsfunc_yy(x_, y_)
        zz = epsfunc_zz(x_, y_)
        xy = epsfunc_xy(x_, y_) if epsfunc_xy is not None else 0 * xx
        yx = epsfunc_yx(x_, y_) if epsfunc_yx is not None else 0 * xx

        return np.array([xx, xy, yx, yy, zz])

    return epsfunc_iso


def create_polygon(x: "np.ndarray", y: "np.ndarray", n: "np.ndarray", detranslate: bool = True) -> list:
    """Given a grid and a refractive index profile, will return the vertices of the polygon for importing into libraries such as Lumerical

    Parameters
    ----------
    x : "np.ndarray"
        the x grid
    y : "np.ndarray"
        the y grid
    n :"np.ndarray"
        the refractive index profile
    detranslate : bool
        if True, will detranslate the vertices

    Returns
    -------
    list[tuples]
        the resulting vertices

    """

    x0, y0 = [x.copy(), y.copy()]
    diff = np.abs(np.diff(n, axis=1))
    where = np.argwhere(diff > np.mean(n))
    tx = np.mean(x[where[:, 0]])
    ty = np.mean(y[where[:, 1]])
    x0 -= tx
    y0 -= ty
    diff = diff[1:, :] + diff[:-1, :]
    diff2 = np.abs(np.diff(n, axis=0))
    diff2 = diff2[:, 1:] + diff2[:, :-1]
    diff = diff + diff2
    diff = np.where(diff, 1, 0)

    newd = {}
    for i in np.argwhere(diff):
        angle = np.angle(x0[i[0]] + 1j * y0[i[1]])
        if detranslate:
            newd[angle] = [y0[i[1]] - ty, x0[i[0]] - tx]
        else:
            newd[angle] = [y0[i[1]], x0[i[0]]]

    od = collections.OrderedDict(sorted(newd.items()))

    if detranslate:
        return np.array(list(od.values())).astype(float)
    else:
        return [np.array(list(od.values())).astype(float), tx, ty]


def interp(
    x: "np.ndarray", y: "np.ndarray", x0: "np.ndarray", y0: "np.ndarray", f: "np.ndarray", centered: bool
) -> "np.ndarray":
    """Interpolate a 2D complex array.

    Parameters
    ----------
    x:"np.ndarray"
        the new x grid
    y:"np.ndarray"
        the new y grid
    x0:"np.ndarray"
        the original x grid
    y0:"np.ndarray"
        the original y grid
    f:"np.ndarray"
        the field to interpolate
    centered:bool
        whether or not it needs to stil be shifted

    Returns
    -------
    np.ndarray
        the interpolated field
    """

    if centered:
        # electric fields and intensity are centered
        x0 = EMpy_gpu.utils.centered1d(x0)
        y0 = EMpy_gpu.utils.centered1d(y0)

    f1r = np.zeros((len(x0), len(y)))
    f1i = np.zeros((len(x0), len(y)))
    for ix0 in range(len(x0)):
        f1r[ix0, :] = np.interp(y, y0, np.real(f[ix0, :]))
        f1i[ix0, :] = np.interp(y, y0, np.imag(f[ix0, :]))
    fr = np.zeros((len(x), len(y)))
    fi = np.zeros((len(x), len(y)))
    for iy in range(len(y)):
        fr[:, iy] = np.interp(x, x0, f1r[:, iy])
        fi[:, iy] = np.interp(x, x0, f1i[:, iy])
    return fr + 1j * fi


def interp1d(x: "np.ndarray", x0: "np.ndarray", f: "np.ndarray", centered: bool) -> "np.ndarray":
    """Interpolate a 1D complex array.

    Parameters
    ----------
    x:"np.ndarray"
        the new grid
    x0:"np.ndarray"
        the original grid
    f:"np.ndarray"
        the field to interpolate
    centered:bool
        whether or not it needs to stil be shifted

    Returns
    -------
    np.ndarray
        the interpolated field
    """

    if centered:
        # electric fields and intensity are centered
        x0 = EMpy_gpu.utils.centered1d(x0)

    f1r = np.interp(x, x0, np.real(f[:]))
    f1i = np.interp(x, x0, np.imag(f[:]))
    return f1r + 1j * f1i


def into_chunks(location: str, name: str, chunk_size: int = 20000000) -> None:
    """Takes a large serialized file and breaks it up into smaller chunk files

    Parameters
    ---------
    location : string
        the absolute or relative path of the large file
    name : string
        the name of the serialized smaller components (will have _chunk_# appended to it)
    chunk_size : int
        how big each save chunk should be
    """
    CHUNK_SIZE = chunk_size
    f = open(location, "rb")
    chunk = f.read(CHUNK_SIZE)
    count = 0
    while chunk:  # loop until the chunk is empty (the file is exhausted)
        with open(name + "_chunk_" + str(count), "wb+") as w:
            w.write(chunk)
        count += 1
        chunk = f.read(CHUNK_SIZE)  # read the next chunk
    f.close()


def from_chunks(location: str, name: str) -> None:
    """Takes a directory of serialized chunks that were made using into_chunks and combines them back into a large serialized file

    Parameters
    ----------
    location : string
        the path of the directory where the chunks are located
    name : string
        the name of the serialized file to create (make sure to include file extension if it matters)
    """
    if location[-1] != "/":
        location += "/"
    f = open(name, "wb+")
    direc = os.listdir(location)
    keys = [int(d[9:]) for d in direc]
    dic = dict(zip(keys, direc))
    for i in sorted(dic):
        f.write(open(location + dic[i], "rb").read())
    f.close()


def _get_eps(
    xc: "np.ndarray", yc: "np.ndarray", epsfunc: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"]
) -> tuple:
    """Used by compute_other_fields and adapted from the EMpy library"""
    tmp = epsfunc(xc, yc)

    def _reshape(tmp):
        """pads the array by duplicating edge values"""
        tmp = np.c_[tmp[:, 0:1], tmp, tmp[:, -1:]]
        tmp = np.r_[tmp[0:1, :], tmp, tmp[-1:, :]]
        return tmp

    if tmp.ndim == 2:
        tmp = _reshape(tmp)
        epsxx = epsyy = epszz = tmp
        epsxy = epsyx = np.zeros_like(epsxx)

    elif tmp.ndim == 3:
        assert tmp.shape[2] == 5, "eps must be NxMx5"
        epsxx = _reshape(tmp[:, :, 0])
        epsxy = _reshape(tmp[:, :, 1])
        epsyx = _reshape(tmp[:, :, 2])
        epsyy = _reshape(tmp[:, :, 3])
        epszz = _reshape(tmp[:, :, 4])

    else:
        raise ValueError("Invalid eps")

    return epsxx, epsxy, epsyx, epsyy, epszz


def compute_other_fields_2D(
    neff: float,
    Hx: "np.ndarray",
    Hy: "np.ndarray",
    wl: float,
    x: "np.ndarray",
    y: "np.ndarray",
    boundary: str,
    epsfunc: Callable[["np.ndarray", "np.ndarray"], "np.ndarray"],
) -> list:
    """Given the Hx and Hy fields, will calculate the rest

    Parameters
    ----------
    neff: float
        the effective index of the EigenMode
    Hx: "np.ndarray"
        the Hx field profile
    Hy: "np.ndarray"
        the Hy field profile
    wl: float
        the wavelength of the simulation
    x: "np.ndarray"
        the x numpy grid array
    y: "np.ndarray"
        the y numpy grid array
    boundary: str
        a string representing the boundary conditions as defined for the electromagneticpython library
    epsfunc: function
        a function of (x,y) that can be called to return the refractive index field profile for the domain

    Returns
    -------
    list
        a list of the field profile np.ndarrays [Hx,Hy,Hz,Ex,Ey,Ez]
    """

    dx = np.diff(x)
    dy = np.diff(y)

    dx = np.r_[dx[0], dx, dx[-1]].reshape(-1, 1)
    dy = np.r_[dy[0], dy, dy[-1]].reshape(1, -1)

    xc = (x[:-1] + x[1:]) / 2
    yc = (y[:-1] + y[1:]) / 2
    epsxx, epsxy, epsyx, epsyy, epszz = _get_eps(xc, yc, epsfunc)

    nx = len(x)
    ny = len(y)

    k = 2 * np.pi / wl

    ones_nx = np.ones((nx, 1))
    ones_ny = np.ones((1, ny))

    n = np.dot(ones_nx, dy[:, 1:]).flatten()
    s = np.dot(ones_nx, dy[:, :-1]).flatten()
    e = np.dot(dx[1:, :], ones_ny).flatten()
    w = np.dot(dx[:-1, :], ones_ny).flatten()

    exx1 = epsxx[:-1, 1:].flatten()
    exx2 = epsxx[:-1, :-1].flatten()
    exx3 = epsxx[1:, :-1].flatten()
    exx4 = epsxx[1:, 1:].flatten()

    eyy1 = epsyy[:-1, 1:].flatten()
    eyy2 = epsyy[:-1, :-1].flatten()
    eyy3 = epsyy[1:, :-1].flatten()
    eyy4 = epsyy[1:, 1:].flatten()

    exy1 = epsxy[:-1, 1:].flatten()
    exy2 = epsxy[:-1, :-1].flatten()
    exy3 = epsxy[1:, :-1].flatten()
    exy4 = epsxy[1:, 1:].flatten()

    eyx1 = epsyx[:-1, 1:].flatten()
    eyx2 = epsyx[:-1, :-1].flatten()
    eyx3 = epsyx[1:, :-1].flatten()
    eyx4 = epsyx[1:, 1:].flatten()

    ezz1 = epszz[:-1, 1:].flatten()
    ezz2 = epszz[:-1, :-1].flatten()
    ezz3 = epszz[1:, :-1].flatten()
    ezz4 = epszz[1:, 1:].flatten()

    b = neff * k

    bzxne = (
        0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * eyx4
        / ezz4
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy3
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * (1 - exx4 / ezz4)
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx1
        * s
    ) / b

    bzxse = (
        -0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * eyx3
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * (1 - exx3 / ezz3)
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * n
        * exx1
        * exx4
    ) / b

    bzxnw = (
        -0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * eyx1
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy2
        * e
        - 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * (1 - exx1 / ezz1)
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx4
        * s
    ) / b

    bzxsw = (
        0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * eyx2
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * e
        - 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * (1 - exx2 / ezz2)
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx3
        * n
        * exx1
        * exx4
    ) / b

    bzxn = (
        (
            0.5
            * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * n
            * ezz1
            * ezz2
            / eyy1
            * (2 * eyy1 / ezz1 / n ** 2 + eyx1 / ezz1 / n / w)
            + 0.5
            * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * n
            * ezz4
            * ezz3
            / eyy4
            * (2 * eyy4 / ezz4 / n ** 2 - eyx4 / ezz4 / n / e)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                0.5
                * ezz4
                * ((1 - exx1 / ezz1) / n / w - exy1 / ezz1 * (2.0 / n ** 2 - 2 / n ** 2 * s / (n + s)))
                / exx1
                * ezz1
                * w
                + (ezz4 - ezz1) * s / n / (n + s)
                + 0.5
                * ezz1
                * (-(1 - exx4 / ezz4) / n / e - exy4 / ezz4 * (2.0 / n ** 2 - 2 / n ** 2 * s / (n + s)))
                / exx4
                * ezz4
                * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                -ezz3 * exy2 / n / (n + s) / exx2 * w
                + (ezz3 - ezz2) * s / n / (n + s)
                - ezz2 * exy3 / n / (n + s) / exx3 * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxs = (
        (
            0.5
            * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * s
            * ezz2
            * ezz1
            / eyy2
            * (2 * eyy2 / ezz2 / s ** 2 - eyx2 / ezz2 / s / w)
            + 0.5
            * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * s
            * ezz3
            * ezz4
            / eyy3
            * (2 * eyy3 / ezz3 / s ** 2 + eyx3 / ezz3 / s / e)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                -ezz4 * exy1 / s / (n + s) / exx1 * w
                - (ezz4 - ezz1) * n / s / (n + s)
                - ezz1 * exy4 / s / (n + s) / exx4 * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                0.5
                * ezz3
                * (-(1 - exx2 / ezz2) / s / w - exy2 / ezz2 * (2.0 / s ** 2 - 2 / s ** 2 * n / (n + s)))
                / exx2
                * ezz2
                * w
                - (ezz3 - ezz2) * n / s / (n + s)
                + 0.5
                * ezz2
                * ((1 - exx3 / ezz3) / s / e - exy3 / ezz3 * (2.0 / s ** 2 - 2 / s ** 2 * n / (n + s)))
                / exx3
                * ezz3
                * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxe = (
        (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * (
            0.5 * n * ezz4 * ezz3 / eyy4 * (2.0 / e ** 2 - eyx4 / ezz4 / n / e)
            + 0.5 * s * ezz3 * ezz4 / eyy3 * (2.0 / e ** 2 + eyx3 / ezz3 / s / e)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            -0.5 * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e) * ezz1 * (1 - exx4 / ezz4) / n / exx4 * ezz4
            - 0.5 * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e) * ezz2 * (1 - exx3 / ezz3) / s / exx3 * ezz3
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxw = (
        (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * (
            0.5 * n * ezz1 * ezz2 / eyy1 * (2.0 / w ** 2 + eyx1 / ezz1 / n / w)
            + 0.5 * s * ezz2 * ezz1 / eyy2 * (2.0 / w ** 2 - eyx2 / ezz2 / s / w)
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            0.5 * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e) * ezz4 * (1 - exx1 / ezz1) / n / exx1 * ezz1
            + 0.5 * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e) * ezz3 * (1 - exx2 / ezz2) / s / exx2 * ezz2
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzxp = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                0.5
                * n
                * ezz1
                * ezz2
                / eyy1
                * (-2.0 / w ** 2 - 2 * eyy1 / ezz1 / n ** 2 + k ** 2 * eyy1 - eyx1 / ezz1 / n / w)
                + 0.5
                * s
                * ezz2
                * ezz1
                / eyy2
                * (-2.0 / w ** 2 - 2 * eyy2 / ezz2 / s ** 2 + k ** 2 * eyy2 + eyx2 / ezz2 / s / w)
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                0.5
                * n
                * ezz4
                * ezz3
                / eyy4
                * (-2.0 / e ** 2 - 2 * eyy4 / ezz4 / n ** 2 + k ** 2 * eyy4 + eyx4 / ezz4 / n / e)
                + 0.5
                * s
                * ezz3
                * ezz4
                / eyy3
                * (-2.0 / e ** 2 - 2 * eyy3 / ezz3 / s ** 2 + k ** 2 * eyy3 - eyx3 / ezz3 / s / e)
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                0.5
                * ezz4
                * (
                    -(k ** 2) * exy1
                    - (1 - exx1 / ezz1) / n / w
                    - exy1 / ezz1 * (-2.0 / n ** 2 - 2 / n ** 2 * (n - s) / s)
                )
                / exx1
                * ezz1
                * w
                + (ezz4 - ezz1) * (n - s) / n / s
                + 0.5
                * ezz1
                * (
                    -(k ** 2) * exy4
                    + (1 - exx4 / ezz4) / n / e
                    - exy4 / ezz4 * (-2.0 / n ** 2 - 2 / n ** 2 * (n - s) / s)
                )
                / exx4
                * ezz4
                * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                0.5
                * ezz3
                * (
                    -(k ** 2) * exy2
                    + (1 - exx2 / ezz2) / s / w
                    - exy2 / ezz2 * (-2.0 / s ** 2 + 2 / s ** 2 * (n - s) / n)
                )
                / exx2
                * ezz2
                * w
                + (ezz3 - ezz2) * (n - s) / n / s
                + 0.5
                * ezz2
                * (
                    -(k ** 2) * exy3
                    - (1 - exx3 / ezz3) / s / e
                    - exy3 / ezz3 * (-2.0 / s ** 2 + 2 / s ** 2 * (n - s) / n)
                )
                / exx3
                * ezz3
                * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzyne = (
        0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * (1 - eyy4 / ezz4)
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy3
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * exy4
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx1
        * s
    ) / b

    bzyse = (
        -0.5
        * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
        * (1 - eyy3 / ezz3)
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy1
        * w
        * eyy2
        + 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * exy3
        / ezz3
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * n
        * exx1
        * exx4
    ) / b

    bzynw = (
        -0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * (1 - eyy1 / ezz1)
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy2
        * e
        - 0.5
        * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * exy1
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * exx4
        * s
    ) / b

    bzysw = (
        0.5
        * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
        * (1 - eyy2 / ezz2)
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * e
        - 0.5
        * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * exy2
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx3
        * n
        * exx1
        * exx4
    ) / b

    bzyn = (
        (
            0.5 * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3) * ezz1 * ezz2 / eyy1 * (1 - eyy1 / ezz1) / w
            - 0.5 * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2) * ezz4 * ezz3 / eyy4 * (1 - eyy4 / ezz4) / e
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
        * (
            0.5 * ezz4 * (2.0 / n ** 2 + exy1 / ezz1 / n / w) / exx1 * ezz1 * w
            + 0.5 * ezz1 * (2.0 / n ** 2 - exy4 / ezz4 / n / e) / exx4 * ezz4 * e
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzys = (
        (
            -0.5 * (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3) * ezz2 * ezz1 / eyy2 * (1 - eyy2 / ezz2) / w
            + 0.5 * (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2) * ezz3 * ezz4 / eyy3 * (1 - eyy3 / ezz3) / e
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
        * (
            0.5 * ezz3 * (2.0 / s ** 2 - exy2 / ezz2 / s / w) / exx2 * ezz2 * w
            + 0.5 * ezz2 * (2.0 / s ** 2 + exy3 / ezz3 / s / e) / exx3 * ezz3 * e
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzye = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                -n * ezz2 / eyy1 * eyx1 / e / (e + w)
                + (ezz1 - ezz2) * w / e / (e + w)
                - s * ezz1 / eyy2 * eyx2 / e / (e + w)
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                0.5
                * n
                * ezz4
                * ezz3
                / eyy4
                * (-(1 - eyy4 / ezz4) / n / e - eyx4 / ezz4 * (2.0 / e ** 2 - 2 / e ** 2 * w / (e + w)))
                + 0.5
                * s
                * ezz3
                * ezz4
                / eyy3
                * ((1 - eyy3 / ezz3) / s / e - eyx3 / ezz3 * (2.0 / e ** 2 - 2 / e ** 2 * w / (e + w)))
                + (ezz4 - ezz3) * w / e / (e + w)
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            0.5
            * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * ezz1
            * (2 * exx4 / ezz4 / e ** 2 - exy4 / ezz4 / n / e)
            / exx4
            * ezz4
            * e
            - 0.5
            * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * ezz2
            * (2 * exx3 / ezz3 / e ** 2 + exy3 / ezz3 / s / e)
            / exx3
            * ezz3
            * e
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzyw = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                0.5
                * n
                * ezz1
                * ezz2
                / eyy1
                * ((1 - eyy1 / ezz1) / n / w - eyx1 / ezz1 * (2.0 / w ** 2 - 2 / w ** 2 * e / (e + w)))
                - (ezz1 - ezz2) * e / w / (e + w)
                + 0.5
                * s
                * ezz2
                * ezz1
                / eyy2
                * (-(1 - eyy2 / ezz2) / s / w - eyx2 / ezz2 * (2.0 / w ** 2 - 2 / w ** 2 * e / (e + w)))
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                -n * ezz3 / eyy4 * eyx4 / w / (e + w)
                - s * ezz4 / eyy3 * eyx3 / w / (e + w)
                - (ezz4 - ezz3) * e / w / (e + w)
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            0.5
            * (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * ezz4
            * (2 * exx1 / ezz1 / w ** 2 + exy1 / ezz1 / n / w)
            / exx1
            * ezz1
            * w
            - 0.5
            * (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * ezz3
            * (2 * exx2 / ezz2 / w ** 2 - exy2 / ezz2 / s / w)
            / exx2
            * ezz2
            * w
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    bzyp = (
        (
            (-n * ezz4 * ezz3 / eyy4 - s * ezz3 * ezz4 / eyy3)
            * (
                0.5
                * n
                * ezz1
                * ezz2
                / eyy1
                * (
                    -(k ** 2) * eyx1
                    - (1 - eyy1 / ezz1) / n / w
                    - eyx1 / ezz1 * (-2.0 / w ** 2 + 2 / w ** 2 * (e - w) / e)
                )
                + (ezz1 - ezz2) * (e - w) / e / w
                + 0.5
                * s
                * ezz2
                * ezz1
                / eyy2
                * (
                    -(k ** 2) * eyx2
                    + (1 - eyy2 / ezz2) / s / w
                    - eyx2 / ezz2 * (-2.0 / w ** 2 + 2 / w ** 2 * (e - w) / e)
                )
            )
            + (n * ezz1 * ezz2 / eyy1 + s * ezz2 * ezz1 / eyy2)
            * (
                0.5
                * n
                * ezz4
                * ezz3
                / eyy4
                * (
                    -(k ** 2) * eyx4
                    + (1 - eyy4 / ezz4) / n / e
                    - eyx4 / ezz4 * (-2.0 / e ** 2 - 2 / e ** 2 * (e - w) / w)
                )
                + 0.5
                * s
                * ezz3
                * ezz4
                / eyy3
                * (
                    -(k ** 2) * eyx3
                    - (1 - eyy3 / ezz3) / s / e
                    - eyx3 / ezz3 * (-2.0 / e ** 2 - 2 / e ** 2 * (e - w) / w)
                )
                + (ezz4 - ezz3) * (e - w) / e / w
            )
        )
        / ezz4
        / ezz3
        / (n * eyy3 + s * eyy4)
        / ezz2
        / ezz1
        / (n * eyy2 + s * eyy1)
        / (e + w)
        * eyy4
        * eyy3
        * eyy1
        * w
        * eyy2
        * e
        + (
            (ezz3 / exx2 * ezz2 * w + ezz2 / exx3 * ezz3 * e)
            * (
                0.5
                * ezz4
                * (-2.0 / n ** 2 - 2 * exx1 / ezz1 / w ** 2 + k ** 2 * exx1 - exy1 / ezz1 / n / w)
                / exx1
                * ezz1
                * w
                + 0.5
                * ezz1
                * (-2.0 / n ** 2 - 2 * exx4 / ezz4 / e ** 2 + k ** 2 * exx4 + exy4 / ezz4 / n / e)
                / exx4
                * ezz4
                * e
            )
            - (ezz4 / exx1 * ezz1 * w + ezz1 / exx4 * ezz4 * e)
            * (
                0.5
                * ezz3
                * (-2.0 / s ** 2 - 2 * exx2 / ezz2 / w ** 2 + k ** 2 * exx2 + exy2 / ezz2 / s / w)
                / exx2
                * ezz2
                * w
                + 0.5
                * ezz2
                * (-2.0 / s ** 2 - 2 * exx3 / ezz3 / e ** 2 + k ** 2 * exx3 - exy3 / ezz3 / s / e)
                / exx3
                * ezz3
                * e
            )
        )
        / ezz3
        / ezz2
        / (w * exx3 + e * exx2)
        / ezz4
        / ezz1
        / (w * exx4 + e * exx1)
        / (n + s)
        * exx2
        * exx3
        * n
        * exx1
        * exx4
        * s
    ) / b

    ii = np.arange(nx * ny).reshape(nx, ny)

    # NORTH boundary

    ib = ii[:, -1]

    if boundary[0] == "S":
        sign = 1
    elif boundary[0] == "A":
        sign = -1
    elif boundary[0] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxs[ib] += sign * bzxn[ib]
    bzxse[ib] += sign * bzxne[ib]
    bzxsw[ib] += sign * bzxnw[ib]
    bzys[ib] -= sign * bzyn[ib]
    bzyse[ib] -= sign * bzyne[ib]
    bzysw[ib] -= sign * bzynw[ib]

    # SOUTH boundary

    ib = ii[:, 0]

    if boundary[1] == "S":
        sign = 1
    elif boundary[1] == "A":
        sign = -1
    elif boundary[1] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxn[ib] += sign * bzxs[ib]
    bzxne[ib] += sign * bzxse[ib]
    bzxnw[ib] += sign * bzxsw[ib]
    bzyn[ib] -= sign * bzys[ib]
    bzyne[ib] -= sign * bzyse[ib]
    bzynw[ib] -= sign * bzysw[ib]

    # EAST boundary

    ib = ii[-1, :]

    if boundary[2] == "S":
        sign = 1
    elif boundary[2] == "A":
        sign = -1
    elif boundary[2] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxw[ib] += sign * bzxe[ib]
    bzxnw[ib] += sign * bzxne[ib]
    bzxsw[ib] += sign * bzxse[ib]
    bzyw[ib] -= sign * bzye[ib]
    bzynw[ib] -= sign * bzyne[ib]
    bzysw[ib] -= sign * bzyse[ib]

    # WEST boundary

    ib = ii[0, :]

    if boundary[3] == "S":
        sign = 1
    elif boundary[3] == "A":
        sign = -1
    elif boundary[3] == "0":
        sign = 0
    else:
        raise ValueError("unknown boundary conditions")

    bzxe[ib] += sign * bzxw[ib]
    bzxne[ib] += sign * bzxnw[ib]
    bzxse[ib] += sign * bzxsw[ib]
    bzye[ib] -= sign * bzyw[ib]
    bzyne[ib] -= sign * bzynw[ib]
    bzyse[ib] -= sign * bzysw[ib]

    # Assemble sparse matrix

    iall = ii.flatten()
    i_s = ii[:, :-1].flatten()
    i_n = ii[:, 1:].flatten()
    i_e = ii[1:, :].flatten()
    i_w = ii[:-1, :].flatten()
    i_ne = ii[1:, 1:].flatten()
    i_se = ii[1:, :-1].flatten()
    i_sw = ii[:-1, :-1].flatten()
    i_nw = ii[:-1, 1:].flatten()

    Izx = np.r_[iall, i_w, i_e, i_s, i_n, i_ne, i_se, i_sw, i_nw]
    Jzx = np.r_[iall, i_e, i_w, i_n, i_s, i_sw, i_nw, i_ne, i_se]
    Vzx = np.r_[
        bzxp[iall], bzxe[i_w], bzxw[i_e], bzxn[i_s], bzxs[i_n], bzxsw[i_ne], bzxnw[i_se], bzxne[i_sw], bzxse[i_nw]
    ]

    Izy = np.r_[iall, i_w, i_e, i_s, i_n, i_ne, i_se, i_sw, i_nw]
    Jzy = np.r_[iall, i_e, i_w, i_n, i_s, i_sw, i_nw, i_ne, i_se] + nx * ny
    Vzy = np.r_[
        bzyp[iall], bzye[i_w], bzyw[i_e], bzyn[i_s], bzys[i_n], bzysw[i_ne], bzynw[i_se], bzyne[i_sw], bzyse[i_nw]
    ]

    I = np.r_[Izx, Izy]
    J = np.r_[Jzx, Jzy]
    V = np.r_[Vzx, Vzy]
    B = coo_matrix((V, (I, J))).tocsr()

    HxHy = np.r_[Hx, Hy]
    Hz = B * HxHy.ravel() / 1j
    Hz = Hz.reshape(Hx.shape)

    # in xc e yc
    exx = epsxx[1:-1, 1:-1]
    exy = epsxy[1:-1, 1:-1]
    eyx = epsyx[1:-1, 1:-1]
    eyy = epsyy[1:-1, 1:-1]
    ezz = epszz[1:-1, 1:-1]
    edet = exx * eyy - exy * eyx

    h = e.reshape(nx, ny)[:-1, :-1]
    v = n.reshape(nx, ny)[:-1, :-1]

    # in xc e yc
    Dx = neff * EMpy_gpu.utils.centered2d(Hy) + (Hz[:-1, 1:] + Hz[1:, 1:] - Hz[:-1, :-1] - Hz[1:, :-1]) / (2j * k * v)
    Dy = -neff * EMpy_gpu.utils.centered2d(Hx) - (Hz[1:, :-1] + Hz[1:, 1:] - Hz[:-1, 1:] - Hz[:-1, :-1]) / (2j * k * h)
    Dz = (
        (Hy[1:, :-1] + Hy[1:, 1:] - Hy[:-1, 1:] - Hy[:-1, :-1]) / (2 * h)
        - (Hx[:-1, 1:] + Hx[1:, 1:] - Hx[:-1, :-1] - Hx[1:, :-1]) / (2 * v)
    ) / (1j * k)

    Ex = (eyy * Dx - exy * Dy) / edet
    Ey = (exx * Dy - eyx * Dx) / edet
    Ez = Dz / ezz

    return [Hx, Hy, Hz, Ex, Ey, Ez]



import numpy as np
import numpy
from simphony import Model
from simphony.pins import Pin
from simphony.models import Subcircuit
from emepy.mode import EigenMode
from emepy.fd import ModeSolver
from copy import deepcopy


class Layer(object):
    """Layer objects form the building blocks inside of an EME or PeriodicEME. These represent geometric layers of rectangular waveguides that approximate continuous structures."""

    def __init__(self, mode_solver: ModeSolver, num_modes: int, wavelength: float, length: float) -> None:
        """Layer class constructor

        Parameters
        ----------
        mode_solver : Modesolver
            ModeSolver object used to solve for the modes
        num_modes : int
            Number of total modes for the layer.
        wavelength : number
            Wavelength of eigenmode to solve for (m).
        length : number
            Geometric length of the Layer (m). The length affects the phase of the eigenmodes inside the layer via the complex phasor $e^(jβz)$.
        """

        self.num_modes = num_modes
        self.mode_solver = mode_solver
        self.wavelength = wavelength
        self.length = length
        self.activated_layers = []

    def begin_activate(self):
        return ModelTools._solve_modes_wrapper, self.mode_solver

    def finish_activate(self, sources: list = [], start: float = 0.0, period_length: float = 0.0, mode_solver=None):
        self.mode_solver = mode_solver
        return self.activate_layer(sources, start, period_length, False)

    def activate_layer(
        self, sources: list = [], start: float = 0.0, period_length: float = 0.0, compute_modes=True
    ) -> dict:
        """Solves for the modes in the layer and creates an ActivatedLayer object

        Parameters
        ----------
        sources : list[Source]
            the Sources used to indicate where periodic layers are needed
        start : number
            the starting z value
        periodic_length : number
            the length of a single period

        Returns
        -------
        dict
            a dictionary that maps the period number to the activated layers. If there is no source in a period, it will be None instead at that index

        """

        modes = []

        # Solve for modes
        if compute_modes:
            self.mode_solver.solve()
        for mode in range(self.num_modes):
            modes.append(self.mode_solver.get_mode(mode))

        # Purge spurious mode
        modes = ModelTools.purge_spurious(modes)

        # Create activated layers
        self.activated_layers = dict(zip(sources.keys(), [[] for _ in range(len(sources.keys()))]))

        # Loop through all periods
        for per, srcs in sources.items():

            # Only care about sources between the ends
            start_ = start + per * period_length
            custom_sources = ModelTools.get_sources(srcs, start_, start_ + self.length)

            # First period
            if not per:

                # If no custom sources
                if not len(custom_sources):
                    self.activated_layers[per] += [ActivatedLayer(modes, self.wavelength, self.length)]

                # Other sources
                else:
                    self.activated_layers[per] += ModelTools.get_source_system(
                        modes, self.wavelength, self.length, custom_sources, start_
                    )

            # Any other period
            else:

                # If no custom sources
                if not len(custom_sources):
                    self.activated_layers[per] += [None]

                # Other sources
                else:
                    self.activated_layers[per] += ModelTools.get_source_system(
                        modes, self.wavelength, self.length, custom_sources, start_
                    )

        return self.activated_layers

    def get_activated_layer(self, sources: list = [], start: float = 0.0) -> dict:
        """Gets the activated layer if it exists or calls activate_layer first

        Parameters
        ----------
        sources : list[Source]
            a list of Source objects for this layer

        Returns
        -------
        dict
            a dictionary that maps the period number to the activated layers. If there is no source in a period, it will be None instead at that index
        """

        if not len(self.activated_layers):
            self.activate_layer(sources=sources, start=start)

        return self.activated_layers

    def clear(self) -> "numpy.ndarray":
        """Empties the modes in the ModeSolver to clear memory

        Returns
        -------
        numpy array
            the edited image
        """

        self.mode_solver.clear()


class Duplicator(Model):
    """Duplicator is used for observing scattering parameters in the middle of an arbitrary network"""

    def __init__(self, wavelength: float, num_modes: int, label: str = "", **kwargs) -> None:
        """Creates an instance of Duplicator which is used only for finding the scattering values inside of an arbitrary network after cascaded

        Parameters
        ----------
        wavelength : float
            the wavelength of the simulation
        num_modes : int
            number of modes in the layer being duplicated
        label : str
            The label indicating where this specific duplicator looks
        """
        self.num_modes = num_modes
        self.wavelength = wavelength

        self.left_pins = ["left" + str(i) for i in range(self.num_modes)] + [
            "left_dup{}{}".format(str(i), label) for i in range(self.num_modes)
        ]
        self.right_pins = ["right" + str(i) for i in range(self.num_modes)] + [
            "right_dup{}{}".format(str(i), label) for i in range(self.num_modes)
        ]
        self.S0 = None

        # create the pins for the model
        pins = []
        for name in self.left_pins:
            pins.append(Pin(self, name))
        for name in self.right_pins:
            if "dup" in name:
                pins.append(Pin(self, name))
        for name in self.right_pins:
            if "dup" not in name:
                pins.append(Pin(self, name))
        self.pins = pins
        super().__init__(**kwargs, pins=pins)
        self.s_params = self.calculate_s_params()

    def s_parameters(self, freqs: "np.array" = None) -> "np.ndarray":

        return self.s_params

    def calculate_s_params(self) -> "np.ndarray":
        """Calculates the scattering parameters for the duplicator model"""

        # Create template for final s matrix
        m = self.num_modes

        # Create propagation diagonal matrix
        propagation_matrix1 = np.diag(np.exp((0j) * np.ones(self.num_modes * 4)))

        # Create sub matrix
        s_matrix = np.zeros((2 * m, 2 * m), dtype=complex)
        s_matrix[0:m, 0:m] = propagation_matrix1[m : 2 * m, 0:m]
        s_matrix[m : 2 * m, 0:m] = propagation_matrix1[0:m, 0:m]
        s_matrix[0:m, m : 2 * m] = propagation_matrix1[m : 2 * m, m : 2 * m]
        s_matrix[m : 2 * m, m : 2 * m] = propagation_matrix1[0:m, m : 2 * m]

        # Join all
        s_matrix_new = np.zeros((4 * m, 4 * m), dtype=complex)
        s_matrix_new[:m, 3 * m :] = s_matrix[:m, m:]
        s_matrix_new[m : 2 * m, 3 * m :] = s_matrix[:m, m:]
        s_matrix_new[3 * m :, m : 2 * m] = s_matrix[:m, m:]
        s_matrix_new[:m, 2 * m : 3 * m] = s_matrix[m:, :m]
        s_matrix_new[2 * m : 3 * m, :m] = s_matrix[m:, :m]
        s_matrix_new[3 * m :, :m] = s_matrix[m:, :m]
        s_matrix = s_matrix_new

        # Assign number of ports
        self.right_ports = m  # 2 * m - self.which_s * m
        self.left_ports = m  # 2 * m - (1 - self.which_s) * m
        self.num_ports = 2 * m  # 3 * m
        s_matrix = s_matrix.reshape(1, 4 * m, 4 * m)

        return s_matrix


class Current(Model):
    """The object that the EME uses to track the s_parameters and cascade them as they come along to save memory"""

    def __init__(self, wavelength: float, s: "np.ndarray", **kwargs) -> None:
        """Current class constructor

        Parameters
        ----------
        wavelength : number
            the wavelength of the simulation
        s : numpy array
            the starting scattering matrix
        """
        self.left_ports = s.left_ports
        self.left_pins = s.left_pins
        self.s_params = s.s_params
        self.right_ports = s.right_ports
        self.num_ports = self.right_ports + self.left_ports
        self.right_pins = s.right_pins
        self.wavelength = wavelength

        # create the pins for the model
        pins = []
        for name in self.left_pins:
            pins.append(Pin(self, name))
        for name in self.right_pins:
            pins.append(Pin(self, name))

        super().__init__(**kwargs, pins=pins)

    def update_s(self, s: "np.ndarray", layer: "Layer") -> None:
        """Updates the scattering matrix of the object

        Parameters
        ----------
        s : numpy array
            scattering matrix to use as the update
        layer : Layer
            the layer object whos ports to match
        """

        self.s_params = s
        self.right_ports = layer.right_ports
        self.num_ports = self.right_ports + self.left_ports
        self.right_pins = layer.right_pins

        # create the pins for the model
        pins = []
        for name in self.left_pins:
            pins.append(Pin(self, name))
        for name in self.right_pins:
            pins.append(Pin(self, name))

        super().__init__(pins=pins)

    def s_parameters(self, freqs: "np.ndarray" = None) -> "np.ndarray":
        """Returns the scattering matrix.

        Returns
        -------
        numpy array
            the scattering matrix
        """
        return self.s_params


class ActivatedLayer(Model):
    """ActivatedLayer is produced by the Layer class after the ModeSolvers calculate eigenmodes. This is used to create interfaces. This inherits from Simphony's Model class."""

    def __init__(self, modes: list, wavelength: float, length: float, n_only: bool = False, **kwargs) -> None:
        """ActivatedLayer class constructor

        Parameters
        ----------
        modes : list [Mode]
            list of solved eigenmodes in Mode class form
        wavelength : number
            the wavelength of the eigenmodes
        length : number
            the length of the layer object that produced the eigenmodes. This number is used for phase propagation.
        n_only : bool
            if true, will only use the refractive index profile (default: False)
        """

        self.num_modes = len(modes)
        self.modes = modes
        self.wavelength = wavelength
        self.length = length
        self.left_pins = ["left" + str(i) for i in range(self.num_modes)]
        self.right_pins = ["right" + str(i) for i in range(self.num_modes)]
        self.S0 = None
        self.nk = []
        self.pk = []
        if not n_only:
            self.normalize_fields()
            self.s_params = self.calculate_s_params()

        # create the pins for the model
        pins = []
        for name in self.left_pins:
            pins.append(Pin(self, name))
        for name in self.right_pins:
            pins.append(Pin(self, name))

        self.pins = pins
        super().__init__(**kwargs, pins=pins)

    def normalize_fields(self) -> None:
        """Normalizes all of the eigenmodes such that the overlap with its self, power, is 1."""

        for mode in range(len(self.modes)):
            self.modes[mode].normalize()

    def calculate_s_params(self) -> "np.ndarray":
        """Calculates the s params for the phase propagation and returns it.

        Returns
        -------
        numpy array
            the scattering matrix for phase propagation.
        """

        # Create template for final s matrix
        m = self.num_modes
        s_matrix = np.zeros((1, 2 * m, 2 * m), dtype=complex)

        # Create eigenvalue vector
        eigenvalues1 = (2 * np.pi) * np.array([mode.neff for mode in self.modes * 2]) / (self.wavelength)

        # Create propagation diagonal matrix
        propagation_matrix1 = np.diag(np.exp(self.length * 1j * eigenvalues1))

        # Assign prop matrix to final s params (moving corners to right spots)
        s_matrix[0, 0:m, 0:m] = propagation_matrix1[m : 2 * m, 0:m]
        s_matrix[0, m : 2 * m, 0:m] = propagation_matrix1[0:m, 0:m]
        s_matrix[0, 0:m, m : 2 * m] = propagation_matrix1[m : 2 * m, m : 2 * m]
        s_matrix[0, m : 2 * m, m : 2 * m] = propagation_matrix1[0:m, m : 2 * m]

        # Assign ports
        self.right_ports = m
        self.left_ports = m
        self.num_ports = 2 * m

        return s_matrix

    def s_parameters(self, freqs: "np.ndarray" = None) -> "np.ndarray":

        return self.s_params


class InterfaceSingleMode(Model):
    """The InterfaceSingleMode class represents the interface between two different layers. This class is an approximation to speed up the process and can ONLY be used during single mode EME."""

    def __init__(self, layer1: "Layer", layer2: "Layer", **kwargs) -> None:
        """InterfaceSingleMode class constructor

        Parameters
        ----------
        layer1 : Layer
            the left Layer object of the interface
        layer2 : Layer
            the right Layer object of the interface
        """

        self.layer1 = layer1
        self.layer2 = layer2
        self.num_modes = 1
        self.left_ports = 1
        self.right_ports = 1
        self.num_ports = self.left_ports + self.right_ports
        self.left_pins = ["left" + str(i) for i in range(self.left_ports)]
        self.right_pins = ["right" + str(i) for i in range(self.right_ports)]

        # create the pins for the model
        pins = []
        for name in self.left_pins:
            pins.append(Pin(self, name))
        for name in self.right_pins:
            pins.append(Pin(self, name))
        super().__init__(**kwargs, pins=pins)
        self.solve()

    def s_parameters(self, freqs: "np.ndarray" = None) -> "np.ndarray":
        """Returns the scattering matrix.

        Returns
        -------
        numpy array
            the scattering matrix
        """

        return self.s_params

    def solve(self) -> None:
        """Solves for the scattering matrix based on transmission and reflection"""

        s = np.zeros((2 * self.num_modes, 2 * self.num_modes), dtype=complex)

        for inp in range(len(self.layer1.modes)):
            for outp in range(len(self.layer2.modes)):

                left_mode = self.layer1.modes[inp]
                right_mode = self.layer2.modes[outp]

                r, t = self.get_values(left_mode, right_mode)

                s[outp, inp] = r
                s[outp + self.num_modes, inp] = t

        for inp in range(len(self.layer2.modes)):
            for outp in range(len(self.layer1.modes)):

                left_mode = self.layer1.modes[outp]
                right_mode = self.layer2.modes[inp]

                r, t = self.get_values(right_mode, left_mode)

                s[outp, inp + self.num_modes] = t
                s[outp + self.num_modes, inp + self.num_modes] = r

        self.s_params = s.reshape((1, 2 * self.num_modes, 2 * self.num_modes))

    def get_values(self, left: EigenMode, right: EigenMode) -> tuple:
        """Returns the reflection and transmission coefficient based on the two modes

        Parameters
        ----------
        left : EigenMode
            leftside eigenmode
        right : EigenMode
            rightside eigenmode

        Returns
        -------
        r : number
            reflection coefficient
        t : number
            transmission coefficient
        """

        a = 0.5 * left.inner_product(right) + 0.5 * right.inner_product(left)
        b = 0.5 * left.inner_product(right) - 0.5 * right.inner_product(left)

        t = (a ** 2 - b ** 2) / a
        r = 1 - t / (a + b)

        return -r, t

    def clear(self) -> None:
        """Clears the scattering matrix in the object"""

        self.s_params = None


class InterfaceMultiMode(Model):
    """The InterfaceMultiMode class represents the interface between two different layers."""

    def __init__(self, layer1: "Layer", layer2: "Layer", **kwargs) -> None:
        """InterfaceMultiMode class constructor

        Parameters
        ----------
        layer1 : Layer
            the left Layer object of the interface
        layer2 : Layer
            the right Layer object of the interface
        """

        self.layer1 = layer1
        self.layer2 = layer2
        self.left_ports = layer1.right_ports
        self.right_ports = layer2.left_ports
        self.left_pins = ["left" + str(i) for i in range(self.left_ports)]
        self.right_pins = ["right" + str(i) for i in range(self.right_ports)]

        # create the pins for the model
        pins = []
        for name in self.left_pins:
            pins.append(Pin(self, name))
        for name in self.right_pins:
            pins.append(Pin(self, name))

        super().__init__(**kwargs, pins=pins)
        self.num_ports = layer1.right_ports + layer2.left_ports
        self.solve()

    def s_parameters(self, freqs: "np.ndarray" = None) -> "np.ndarray":

        return self.s_params

    def solve(self) -> None:
        """Solves for the scattering matrix based on transmission and reflection"""

        s = np.zeros((self.num_ports, self.num_ports), dtype=complex)

        # Forward values
        for p in range(self.left_ports):

            ts = self.get_t(p, self.layer1, self.layer2, self.left_ports)
            rs = self.get_r(p, ts, self.layer1, self.layer2, self.left_ports)

            for t in range(len(ts)):
                s[self.left_ports + t][p] = ts[t]
            for r in range(len(rs)):
                s[r][p] = rs[r]

        # Reverse values
        for p in range(self.right_ports):

            ts = self.get_t(p, self.layer2, self.layer1, self.right_ports)
            rs = self.get_r(p, ts, self.layer2, self.layer1, self.right_ports)

            for t in range(len(ts)):
                s[t][self.left_ports + p] = ts[t]
            for r in range(len(rs)):
                s[self.left_ports + r][self.left_ports + p] = rs[r]

        # Keep s params and clear the layers
        self.s_params = s.reshape((1, self.num_ports, self.num_ports))
        self.layer1 = None
        self.layer2 = None

    def get_t(self, p: int, left: "EigenMode", right: "EigenMode", curr_ports: int) -> "np.ndarray":
        """Returns the transmission coefficient based on the two modes

        Parameters
        ----------
        p : int
            port number to look at
        left : Mode
            leftside eigenmode
        right : Mode
            rightside eigenmode
        curr_ports : int
            total number of ports

        Returns
        -------
        np.ndarray
            transmission coefficients
        """

        # Ax = b
        A = np.array(
            [
                [
                    right.modes[k].inner_product(left.modes[i]) + left.modes[i].inner_product(right.modes[k])
                    for k in range(self.num_ports - curr_ports)
                ]
                for i in range(curr_ports)
            ]
        )
        b = np.array([0 if i != p else 2 * left.modes[p].inner_product(left.modes[p]) for i in range(curr_ports)])
        x = np.matmul(np.linalg.pinv(A), b)

        return x

    def get_r(self, p: int, x: "np.ndarray", left: EigenMode, right: EigenMode, curr_ports: int) -> "np.ndarray":
        """Returns the transmission coefficient based on the two modes

        Parameters
        ----------
        p : int
            port number to look at
        x : np.ndarray
            transmission coefficients
        left : Mode
            leftside eigenmode
        right : Mode
            rightside eigenmode
        curr_ports : int
            total number of ports

        Returns
        -------
        r : number
            reflection coefficient
        """

        rs = np.array(
            [
                np.sum(
                    [
                        (right.modes[k].inner_product(left.modes[i]) - left.modes[i].inner_product(right.modes[k]))
                        * x[k]
                        for k in range(self.num_ports - curr_ports)
                    ]
                )
                / (2 * left.modes[i].inner_product(left.modes[i]))
                for i in range(curr_ports)
            ]
        )

        return rs

    def clear(self) -> None:
        """Clears the scattering matrix in the object"""

        self.s_params = None


class SourceDuplicator(Model):
    """SourceDuplicator is used for custom sources at an arbitrary location inside of the network"""

    def __init__(
        self,
        wavelength: float,
        modes: list,
        length: float,
        pk: list = [],
        nk: list = [],
        label: str = "",
        special_left: list = [],
        special_right: list = [],
        **kwargs,
    ) -> None:
        """Like Duplicator, but is used as a custom input rather than peaker. Optimized for minimum ports necessary

        Parameters
        ----------
        wavelength : number
            wavelength of the simulation
        modes : list[EigenMode]
            list of the EigenModes at the layer being duplicated
        length : float
            the length of the duplicated layer
        pk : list[float]
            a list of the mode coefficients propagating in the positive direction
        nk : list[float]
            a list of the mode coefficients propagating in the negative direction
        label : str
            the label indicating the location of the duplicator
        special_left : list
            the left coefficients to keep that are not included in pk
        special_right : list
            the right coefficients to keep taht are not included in nk
        """

        self.num_modes = len(modes)
        self.wavelength = wavelength
        self.modes = modes
        self.length = length
        self.pk = pk
        self.nk = nk
        self.normalize_fields()

        self.left_pins = (
            ["left" + str(i) for i in range(self.num_modes)]
            + ["left_dup{}{}".format(str(i), label) for i in range(len(pk) * (not len(special_left)))]
            + special_left
        )
        self.right_pins = (
            ["right" + str(i) for i in range(self.num_modes)]
            + ["right_dup{}{}".format(str(i), label) for i in range(len(nk) * (not len(special_right)))]
            + special_right
        )
        self.S0 = None
        self.S1 = None

        # create the pins for the model
        pins = []
        for name in self.left_pins:
            pins.append(Pin(self, name))
        for name in self.right_pins:
            if "dup" in name:
                pins.append(Pin(self, name))
        for name in self.right_pins:
            if "dup" not in name:
                pins.append(Pin(self, name))
        self.pins = pins
        super().__init__(**kwargs, pins=pins)
        self.s_params = self.calculate_s_params()

    def s_parameters(self, freqs: "np.array" = None) -> "np.ndarray":

        return self.s_params

    def normalize_fields(self) -> None:
        """Normalizes all of the eigenmodes such that the overlap with its self, power, is 1."""

        for mode in range(len(self.modes)):
            self.modes[mode].normalize()

    def calculate_s_params(self) -> "np.ndarray":
        """Calculates the scattering parameters for the system"""

        # Create template for final s matrix
        m = self.num_modes

        # Create eigenvalue vector
        eigenvalues1 = (2 * np.pi) * np.array([mode.neff for mode in self.modes * 4]) / (self.wavelength)

        # Create propagation diagonal matrix
        propagation_matrix1 = np.diag(np.exp(self.length * 1j * eigenvalues1))

        # Create sub matrix
        s_matrix = np.zeros((2 * m, 2 * m), dtype=complex)
        s_matrix[0:m, 0:m] = propagation_matrix1[m : 2 * m, 0:m]
        s_matrix[m : 2 * m, 0:m] = propagation_matrix1[0:m, 0:m]
        s_matrix[0:m, m : 2 * m] = propagation_matrix1[m : 2 * m, m : 2 * m]
        s_matrix[m : 2 * m, m : 2 * m] = propagation_matrix1[0:m, m : 2 * m]

        # Join all
        s_matrix_new = np.zeros((4 * m, 4 * m), dtype=complex)
        s_matrix_new[:m, 3 * m :] = s_matrix[:m, m:]
        s_matrix_new[m : 2 * m, 3 * m :] = s_matrix[:m, m:]
        s_matrix_new[3 * m :, m : 2 * m] = s_matrix[:m, m:]
        s_matrix_new[:m, 2 * m : 3 * m] = s_matrix[m:, :m]
        s_matrix_new[2 * m : 3 * m, :m] = s_matrix[m:, :m]
        s_matrix_new[3 * m :, :m] = s_matrix[m:, :m]
        # s_matrix_new[m:3*m,m:3*m] = s_matrix[:,:]
        s_matrix = s_matrix_new

        # Delete rows and cols
        pk_remove = m - len(self.pk)
        nk_remove = m - len(self.nk)
        for _ in range(nk_remove):
            s_matrix = np.delete(s_matrix, -m - 1, axis=0)
            s_matrix = np.delete(s_matrix, -m - 1, axis=1)
        for _ in range(pk_remove):
            s_matrix = np.delete(s_matrix, -m - len(self.nk) - 1, axis=0)
            s_matrix = np.delete(s_matrix, -m - len(self.nk) - 1, axis=1)

        # Assign number of ports
        self.right_ports = m  # 2 * m - self.which_s * m
        self.left_ports = m  # 2 * m - (1 - self.which_s) * m
        self.num_ports = 2 * m  # 3 * m
        s_matrix = s_matrix.reshape(1, 2 * m + len(self.pk) + len(self.nk), 2 * m + len(self.pk) + len(self.nk))

        return s_matrix


class CopyModel(Model):
    """A simple Model that can be used to deep copy any of EMEPy's Models"""

    def __init__(self, model: "Model", keep_modes: bool = True, **kwargs) -> None:
        """Creates an instance of CopyModel by deepcopying all the attributes of model

        Parameters
        ----------
        model : Model
            the model ot deepcopy
        """

        self.num_modes = model.num_modes if hasattr(model, "num_modes") else None
        self.modes = model.modes if hasattr(model, "modes") and keep_modes else []
        self.wavelength = model.wavelength if hasattr(model, "wavelength") else None
        self.length = model.length if hasattr(model, "length") else None
        self.left_ports = model.left_ports if hasattr(model, "left_ports") else 0
        self.right_ports = model.right_ports if hasattr(model, "right_ports") else 0

        self.S0 = ModelTools.make_copy_model(model.S0, keep_modes=False) if hasattr(model, "S0") else None
        self.nk = model.nk if hasattr(model, "nk") else []
        self.pk = model.pk if hasattr(model, "pk") else []
        self.pins = self.copy_pins(model.pins) if hasattr(model, "pins") else []
        self.left_pins = (
            model.left_pins if hasattr(model, "left_pins") else []
        )  # [self.find_pin(pin) for pin in model.left_pins] if hasattr(model,"left_pins") else []
        self.right_pins = (
            model.right_pins if hasattr(model, "right_pins") else []
        )  # [self.find_pin(pin) for pin in model.right_pins] if hasattr(model,"right_pins") else []
        self.s_params = model.s_parameters([0])

        super().__init__(**kwargs, pins=self.pins)
        return

    def s_parameters(self, freqs: "np.array" = None) -> "np.ndarray":

        return self.s_params

    def copy_pins(self, pins: list) -> list:
        """Copies the pins by creating new pins that are no longer attached to the original Model's connections

        Parameters
        ----------
        pins : list[Pin]
            the list of simphony pins to deepcopy

        Returns
        -------
        list[Pin]
            returns a list of copied pins
        """

        return [Pin(self, p.name) for p in pins]

    def find_pin(self, pin: "Pin") -> "Pin":
        """Searches the instance's pins for the provided pin based on names

        Parameters
        ----------
        pin : Pin
            the pin whos name to search for in the pinlist

        Returns
        -------
        Pin
            pin object corresponding to the searched pin
        """

        for p in self.pins:
            if p.name == pin.name:
                return p
        return Pin(self, pin.name)


class ModelTools(object):
    @staticmethod
    def purge_spurious(modes: list) -> list:
        """Purges all spurious modes in the dataset to prevent EME failure for high mode simulations

        Parameters
        ----------
        modes : list[EigenMode]
            list of EigenModes to remove spurious from

        Returns
        -------
        list[EigenMode]
            returns a list of the eigenmodes that are not spurious
        """

        mm = deepcopy(modes)
        for i, mode in enumerate(mm[::-1]):
            if mode.check_spurious():
                mm.remove(mode)

        return mm

    @staticmethod
    def get_sources(sources: list = [], start: float = 0.0, end: float = 0.0) -> list:
        """Given a list of sources, returns a new list of sources that can be found within the given range

        Parameters
        ----------
        sources : list[Source]
            list of Source objects
        start : number
            the starting point of the range
        end : number
            the ending point of the range

        Returns
        -------
        list[Source]
            returns a list of sources within the provided range
        """

        return [
            i
            for i in sources
            if i.z is not None and (((start <= i.z < end) and i.k) or ((start < i.z <= end) and not i.k))
        ]

    @staticmethod
    def get_source_system(
        modes: list, wavelength: float, length: float, custom_sources: list, start: float = 0.0
    ) -> list:
        """Creates SourceDuplicators matching the provided source locations and modes provided. This is a replacement for creating ActivatedLayers which contain no information about sources.

        Parameters
        ----------
        modes : list[EigenMode]
            the solved modes in the system
        wavelength : float
            the wavelength of the simulation
        length : float
            the length of the layer
        custom_sources : list[Source]
            the custom sources used to create SourceDuplicator objects
        start : float
            the starting z value for these layers

        Returns
        -------
        list[SourceDuplicator]
            returns the newly created source duplicators whom when cascaded form an equivalent ActivatedLayer with bonus source locations
        """

        # Get lengths between components
        lengths = np.diff([start] + [i.z for i in custom_sources] + [start + length])
        dups = []

        # Enumerate through all lengths
        length_tracker = start
        for i, length in enumerate(lengths):

            # Create coefficents indexes to keep in the models
            pk, nk = [[], []]
            if (i - 1) > -1 and custom_sources[i - 1].k:
                pk = custom_sources[i - 1].mode_coeffs
            if (i) < len(custom_sources) and not custom_sources[i].k:
                nk = custom_sources[i].mode_coeffs

            # Create label
            left = custom_sources[i - 1].get_label() if (i - 1) > -1 else "n" + str(length_tracker)
            right = custom_sources[i].get_label() if (i) < len(custom_sources) else "n" + str(length_tracker + length)
            label = "_{}_to_{}".format(left, right)

            # Create duplicators
            dups.append(SourceDuplicator(wavelength, modes, length, pk=pk, nk=nk, label=label))
            length_tracker += length

        return dups

    @staticmethod
    def _prop_all(*args) -> "Model":
        """Given an arbitrary amount of simphony models as inputs, cascades them and returns the cascaded network"""

        layers = [ModelTools.make_copy_model(a, keep_modes=False) for a in args if a is not None]

        temp_s = layers[0]
        for s in layers[1:]:
            Subcircuit.clear_scache()

            # make sure the components are completely disconnected
            temp_s.disconnect()
            s.disconnect()

            # # connect the components
            right_pins = [i for i in temp_s.pins if "dup" not in i.name and "left" not in i.name]
            for port in range(len(right_pins)):
                temp_s[f"right{port}"].connect(s[f"left{port}"])

            temp_s = temp_s.circuit.to_subcircuit()
            temp_s.s_params = temp_s.s_parameters([0])

        return ModelTools.make_copy_model(temp_s, keep_modes=False)

    @staticmethod
    def periodic_duplicate_format(model: "Model", start: float, end: float) -> "Model":
        """Checks if the model that has custom sources installed actually needs the installed sources between start and end. If not, removes those pins and rows/columns from the s_matrices

        Parameters
        ----------
        model : Model
            the simphony model to check
        start : float
            the starting point of the range of concern
        end : float
            the ending point of the range of concern

        Returns
        -------
        Model
            the simplified simphony model
        """

        model = ModelTools.make_copy_model(model)
        wanted = [i for i, pin in enumerate(model.pins) if "dup" in pin.name]
        indices = [i for i, pin in enumerate(model.pins) if "dup" in pin.name]
        model.s_params = model.s_parameters([0])

        # Find wanted indices
        for i in wanted[::-1]:
            if "_to_" not in model.pins[i].name:
                pass
            elif "left" in model.pins[i].name:
                n = model.pins[i].name
                n_ = n.split("_")
                l, _ = (float(n_[2][1:]), float(n_[4][1:]))
                if start <= l < end or np.isclose(start, l, 1e-5):
                    continue
            else:
                n = model.pins[i].name
                n_ = n.split("_")
                _, r = (float(n_[2][1:]), float(n_[4][1:]))
                if start < r <= end or np.isclose(r, end, 1e-5):
                    continue
            wanted.remove(i)

        # Remove unwanted rows and columns
        for i in indices:
            if i not in wanted:
                model.s_params = np.delete(model.s_params, i, 1)
                model.s_params = np.delete(model.s_params, i, 2)

        # Remove unwanted pins
        model.pins = [i for j, i in enumerate(model.pins) if j in wanted or "dup" not in i.name]
        pin_names = [i.name for i in model.pins]
        model.left_pins = [i for i in model.left_pins if i in pin_names]
        model.right_pins = [i for i in model.right_pins if i in pin_names]

        return model

    @staticmethod
    def make_copy_model(model: Model, keep_modes=True) -> CopyModel:
        """Takes an EMEPy model (inheriting a simphony model) and deepcopies it

        Parameters
        ----------
        model: Model
            the EMEPy model to copy

        Returns
        -------
        CopyModel
            the deepcopied model
        """
        new_model = CopyModel(model, keep_modes) if model is not None else None
        return new_model

    @staticmethod
    def compute(model, pin_values: "dict", freq: "float" = 0) -> dict:
        """Takes a dictionary mapping each pin name to a coefficent and multiplies by the S matrix

        Parameters
        ----------
        pin_values : dict
            a mapping of the pin names and the corresponding weights (mode coefficients)
        freq : float
            the frequency to compute at

        Returns
        -------
        dict
           a mapping of the output pins to the resulting weights (mode coefficients)
        """
        cfs = np.zeros(len(model.pins), dtype=complex)
        pin_names = [i.name for i in model.pins]
        key_index = dict(zip(pin_names, [pin_names.index(i) for i in pin_names]))
        for key, value in pin_values.items():
            if key in key_index:
                cfs[key_index[key]] = value
        matrix = model.s_parameters(np.array([freq]))
        output = np.matmul(matrix, cfs)[0]
        return dict(zip(pin_names, output))

    # Define tasks
    @staticmethod
    def layers_task(l, r, interface_type):
        return l, ModelTools._prop_all(l, interface_type(l, r))

    @staticmethod
    def _prop_all_wrapper(arg_list, result_list):
        return ModelTools._prop_all(*arg_list), result_list

    @staticmethod
    def _solve_modes_wrapper(mode_solver):
        mode_solver.solve()
        return mode_solver



class Source(object):
    """This class defines mode sources that can be created for monitors"""

    def __init__(self, z: float = None, mode_coeffs: list = [], k=1) -> None:
        """Constructor for Source. Note: if the modes corresponding to the coefficients defined are not found in the system, the coefficients will be reduced to 0. For now, the wavelength will always be the same as defined in EME.

        Parameters
        ----------
        z : float
            if defined, the z location to define each mode source (default: None)
        mode_coeffs : list[int]
            if z is non-empty, this list represents the modes used for each source. (default: [])
        k : int
            if positive (>0), will propagate in the positive direction
        """

        self.z = z
        self.mode_coeffs = mode_coeffs
        self.k = k > 0

    def get_label(self) -> str:
        """Returns a string that represents the source"""
        k = "+" if self.k else "-"
        return "{}{}".format(k, self.z)

    def match_label(self, label: str) -> bool:
        """Sees if the provided label matches the current Source object's and returns True if so

        Parameters
        ----------
        label : str
            the label to compare

        Returns
        -------
        bool
            True if label matches self

        """
        return label == self.get_label()

    def __str__(self):
        return self.get_label()



from dataclasses import asdict
import numpy as np
from matplotlib import pyplot as plt
from collections.abc import Sequence

from emepy.fd import ModeSolver
from emepy.mode import Mode, Mode1D, EigenMode
from emepy.tools import interp, interp1d, get_epsfunc, rectangle_to_n
from emepy.materials import Si, SiO2
from emepy.geometries import DynamicPolygon
from emepy.eme import Layer

from tidy3d.plugins import ModeSolver as tidy3d_ModeSolver
from tidy3d import PolySlab
import tidy3d as td


class _MSTidy3D(ModeSolver):
    def __init__(
        self,
        wavelength=1.55,
        xz_vertices=None,
        thickness=0.22,
        z_loc=None,
        num_modes=1,
        n_core=None,
        n_cladding=None,
        cladding_width=3.0,
        cladding_thickness=3.0,
        mesh=128,
        x=None,
        y=None,
        target_neff=3.0,
    ):

        # Fix params
        self.wavelength = wavelength
        self.xz_vertices = (
            xz_vertices if xz_vertices is not None else [(0, 0), (0, 1), (1, 1), (1, 0)]
        )
        self.thickness = thickness
        self.z_loc = (
            z_loc
            if z_loc is not None
            else 0.5 * (self.xz_vertices[0][1] + self.xz_vertices[-1][1])
        )
        self.n_core = n_core if n_core is not None else Si(self.wavelength)
        self.n_cladding = (
            n_cladding if n_cladding is not None else SiO2(self.wavelength)
        )
        self.num_modes = num_modes
        self.cladding_width = cladding_width
        self.cladding_thickness = cladding_thickness
        self.target_neff = target_neff

        # Other params
        self.max_z = max([v[1] for v in self.xz_vertices])
        self.min_z = min([v[1] for v in self.xz_vertices])
        self.max_x = max([v[0] for v in self.xz_vertices])
        self.min_x = min([v[0] for v in self.xz_vertices])
        self.delta_z = self.max_z - self.min_z
        self.delta_x = self.max_x - self.min_x
        self.mesh = mesh

        # Get x and y and z
        Lx, Ly, Lz = self.cladding_width, self.cladding_thickness, self.delta_z + 2.0
        self.x = x if x is not None else np.linspace(-Lx / 2, Lx / 2, mesh)
        self.y = y if y is not None else np.linspace(-Ly / 2, Ly / 2, mesh)
        self.z = np.linspace(-Lz / 2, Lz / 2, mesh)
        self.after_x = self.x
        self.after_y = self.y
        self._old_x = np.linspace(-Lx / 2, Lx / 2, mesh + 2)
        self._old_y = np.linspace(-Ly / 2, Ly / 2, mesh + 2)

        self.setup()
        self.n = self.get_any_n(self.x, self.y, self.z_loc)

    def setup(self):
        # Set up geometry
        waveguide = td.Structure(
            geometry=PolySlab(
                vertices=[(z, x) for x, z, in self.xz_vertices],
                axis=2,
                slab_bounds=(-self.thickness / 2, self.thickness / 2),
            ),
            medium=td.Medium(permittivity=self.n_core**2),
        )

        # Set up simulation
        Lx, Ly, Lz = self.delta_z + 2.0, self.cladding_width, self.cladding_thickness

        grid_x = td.CustomGrid(dl=np.diff(self.z).tolist())
        grid_y = td.CustomGrid(dl=np.diff(self.x).tolist())
        grid_z = td.CustomGrid(dl=np.diff(self.y).tolist())
        grid_spec = td.GridSpec(
            grid_x=grid_x, grid_y=grid_y, grid_z=grid_z, wavelength=self.wavelength
        )  # Note the difference in convention for

        self.sim = td.Simulation(
            size=(Lx, Lz, Ly),
            grid_spec=grid_spec,
            structures=[waveguide],
            run_time=1e-12,
            medium=td.Medium(permittivity=self.n_cladding**2),
            boundary_spec=td.BoundarySpec(
                x=td.Boundary.pml(num_layers=20),
                y=td.Boundary.pml(num_layers=20),
                z=td.Boundary.pml(num_layers=20),
            ),
            subpixel=True,
        )

        # Define mode solver
        mode_spec = td.ModeSpec(num_modes=self.num_modes, target_neff=self.target_neff)

        plane = td.Box(center=(self.z_loc, 0, 0), size=(0, Ly, Lz))

        self.mode_solver = tidy3d_ModeSolver(
            simulation=self.sim,
            plane=plane,
            mode_spec=mode_spec,
            freqs=[td.constants.C_0 / self.wavelength],
        )

    def solve(self):

        # Run simulation
        mode_data = self.mode_solver.solve()
        self._old_x = mode_data.sel_mode_index(0).field_components["Ey"].y
        self._old_y = mode_data.sel_mode_index(0).field_components["Ey"].z
        index = self.get_n()
        self.n = index

        # Save modes
        self.modes = []
        for m in range(self.num_modes):
            # Get mode
            mode = mode_data.sel_mode_index(m)

            # Place all fields onto a shared grid
            Ex = mode.field_components["Ey"]
            Ex = interp(
                self.x,
                self.y,
                Ex.y.to_numpy(),
                Ex.z.to_numpy(),
                Ex.to_numpy()[0, :, :, 0],
                centered=False,
            )

            Ey = mode.field_components["Ez"]
            Ey = interp(
                self.x,
                self.y,
                Ey.y.to_numpy(),
                Ey.z.to_numpy(),
                Ey.to_numpy()[0, :, :, 0],
                centered=False,
            )

            Ez = mode.field_components["Ex"]
            Ez = interp(
                self.x,
                self.y,
                Ez.y.to_numpy(),
                Ez.z.to_numpy(),
                Ez.to_numpy()[0, :, :, 0],
                centered=False,
            )

            Hx = mode.field_components["Hy"]
            Hx = interp(
                self.x,
                self.y,
                Hx.y.to_numpy(),
                Hx.z.to_numpy(),
                Hx.to_numpy()[0, :, :, 0],
                centered=False,
            )

            Hy = mode.field_components["Hz"]
            Hy = interp(
                self.x,
                self.y,
                Hy.y.to_numpy(),
                Hy.z.to_numpy(),
                Hy.to_numpy()[0, :, :, 0],
                centered=False,
            )

            Hz = mode.field_components["Hx"]
            Hz = interp(
                self.x,
                self.y,
                Hz.y.to_numpy(),
                Hz.z.to_numpy(),
                Hz.to_numpy()[0, :, :, 0],
                centered=False,
            )

            self.modes.append(
                Mode(
                    x=self.x,
                    y=self.y,
                    neff=mode_data.n_eff.to_numpy()[0][m],
                    Ex=Ex,
                    Ey=Ey,
                    Ez=Ez,
                    Hx=Hx,
                    Hy=Hy,
                    Hz=Hz,
                    n=index,
                )
            )

    def get_mode(self, index):
        return self.modes[index]

    def get_n(self):
        index = np.zeros((self.mesh + 2, self.mesh + 2), dtype=complex)
        index[:, :] = self.mode_solver._solver_eps(td.constants.C_0 / self.wavelength)[
            0, :, :
        ]
        index = interp(self.x, self.y, self._old_x, self._old_y, index, centered=False)
        return np.sqrt(index)

    def get_any_n(self, x=None, y=None, z=None):
        x_flag, y_flag, z_flag = False, False, False
        x = x if x is not None else self.x
        y = y if y is not None else self.y
        z = z if z is not None else self.z
        x = x if isinstance(x, Sequence) or isinstance(x, np.ndarray) else np.array([x])
        y = y if isinstance(y, Sequence) or isinstance(y, np.ndarray) else np.array([y])
        z = z if isinstance(z, Sequence) or isinstance(z, np.ndarray) else np.array([z])

        if len(x.flatten()) == 1:
            x = np.array([x.item(0), x.item(0), x.item(0)])
            x_flag = True
        if len(y.flatten()) == 1:
            y = np.array([y.item(0), y.item(0), y.item(0)])
            y_flag = True
        if len(z.flatten()) == 1:
            z = np.array([z.item(0), z.item(0), z.item(0)])
            z_flag = True

        bound_x = 0.5 * (z[1:] + z[:-1])
        bound_x = np.insert(bound_x, 0, bound_x[0] - np.diff(bound_x)[0])
        bound_x = np.insert(bound_x, len(bound_x), bound_x[-1] + np.diff(bound_x)[-1])

        bound_y = 0.5 * (x[1:] + x[:-1])
        bound_y = np.insert(bound_y, 0, bound_y[0] - np.diff(bound_y)[0])
        bound_y = np.insert(bound_y, len(bound_y), bound_y[-1] + np.diff(bound_y)[-1])

        bound_z = 0.5 * (y[1:] + y[:-1])
        bound_z = np.insert(bound_z, 0, bound_z[0] - np.diff(bound_z)[0])
        bound_z = np.insert(bound_z, len(bound_z), bound_z[-1] + np.diff(bound_z)[-1])

        coords = td.Coords(x=bound_x, y=bound_y, z=bound_z)
        grid = td.Grid(boundaries=coords)
        n = self.sim.epsilon_on_grid(grid).to_numpy().transpose((1, 2, 0))

        if y_flag:
            n = n[:, 0, :]
        if x_flag:
            n = n[0, ...]
        if z_flag:
            n = n[..., 0]

        return np.sqrt(n)


class Tidy3DPoly(DynamicPolygon):
    def __init__(
        self,
        wavelength=1.55,
        xz_vertices=None,
        thickness=0.22,
        num_modes=1,
        n_core=None,
        n_cladding=None,
        cladding_width=3.0,
        cladding_thickness=3.0,
        mesh=128,
        num_layers=1,
    ):
        self.design = xz_vertices
        self.wavelength = wavelength
        self.thickness = thickness
        self.num_modes = num_modes
        self.n_core = n_core
        self.n_cladding = n_cladding
        self.cladding_width = cladding_width
        self.cladding_thickness = cladding_thickness
        self.mesh = mesh
        self.num_layers = num_layers
        self.update()

    def update(self):
        self.mode_solver = _MSTidy3D(
            wavelength=self.wavelength,
            xz_vertices=self.design,
            thickness=self.thickness,
            z_loc=0,
            num_modes=self.num_modes,
            n_core=self.n_core,
            n_cladding=self.n_cladding,
            cladding_width=self.cladding_width,
            cladding_thickness=self.cladding_thickness,
            mesh=self.mesh,
        )

        layer_length = self.mode_solver.delta_z / self.num_layers
        z_loc = lambda i: (i + 0.5) * layer_length + self.mode_solver.min_z
        self.layers = [
            Layer(
                _MSTidy3D(
                    wavelength=self.wavelength,
                    xz_vertices=self.design,
                    thickness=self.thickness,
                    z_loc=z_loc(i),
                    num_modes=self.num_modes,
                    n_core=self.n_core,
                    n_cladding=self.n_cladding,
                    cladding_width=self.cladding_width,
                    cladding_thickness=self.cladding_thickness,
                    mesh=self.mesh,
                ),
                self.num_modes,
                self.wavelength,
                layer_length,
            )
            for i in range(self.num_layers)
        ]

    def get_design(self) -> list:
        """Returns the design region as a list of parameters. Each parameter represents a spacial location in a single direction for a single vertex of the polygon. Note in 2D this would look like [x0,z0,x1,z1,...xn,zn] and in 3D [x0,y0,z0,x1,y1,z1,...xn,yn,zn]"""
        return self.design

    def set_design(self, design) -> list:
        """Sets the design region"""
        self.design = design
        self.update()

    def get_n(self, grid_x, grid_y, grid_z) -> "np.ndarray":
        """Returns the index profile for the grid provided"""
        return self.mode_solver.get_any_n(grid_x, grid_y, grid_z)


if __name__ == "__main__":

    # Define polygon
    xz_vertices = [
        (-0.4, -0.5),
        (0.4, -0.5),
        (0.45, 0.0),
        (0.4, 0.5),
        (-0.4, 0.5),
        (-0.45, 0.0),
    ]
    thickness = 0.22
    wavelength = 1.55
    z_loc = 0.4
    n_core = Si(wavelength)
    n_cladding = SiO2(wavelength)
    num_modes = 4

    # Define solver
    polygon = Tidy3DPoly(
        wavelength,
        xz_vertices,
        thickness,
        num_modes,
        n_core,
        n_cladding,
        mesh=50,
        num_layers=1,
    )

    # xz_vertices = [
    #     (-0.4, -0.5),
    #     (0.4, -0.5),
    #     (0.445, 0.0),
    #     (0.4, 0.5),
    #     (-0.4, 0.5),
    #     (-0.445, 0.0),
    # ]
    # polygon2 = Tidy3DPoly(
    #     wavelength, xz_vertices, thickness, num_modes, n_core, n_cladding, mesh=50, num_layers=1
    # )
    # polygon.layers[0].mode_solver.solve()
    # # polygon2.layers[0].mode_solver.solve()
    # n1 = polygon.layers[0].mode_solver.n.real
    # n2 = polygon2.layers[0].mode_solver.n.real

    # plt.figure()
    # plt.imshow(np.rot90(n2), cmap="Greys")
    # plt.show()

    # from emepy.eme import EME

    # eme = EME(layers=[*polygon])
    # # eme.draw()
    # # plt.show()

    from emepy.fd import MSTidy3D

    solver = MSTidy3D(1.55, 0.5, 0.22, num_modes=5, mesh=100, PML=False)

    solver.solve()

    # solver.get_mode(0).plot_material()
    # plt.show()
    for i in range(5):
        solver.get_mode(i).plot()
        print(solver.get_mode(i).neff)
        plt.show()
